package no.rwr.awt.x;

import java.awt.*;
import java.awt.peer.*;
import java.awt.image.*;
import no.rwr.x.WMSizeHints;
import no.rwr.x.WindowAttributes;
import no.rwr.x.Display;
import no.rwr.x.Visual;
import no.rwr.x.Screen;
import no.rwr.awt.EventConstants;

import java.util.EventListener;
import java.awt.event.MouseListener;



public class XCanvasPeer implements CanvasPeer {

    static final Dimension MIN_SIZE = new Dimension(1, 1);

    no.rwr.x.Window window;
    no.rwr.x.Window parent;

    Component component;
    XGraphicsConfiguration config;

    public XCanvasPeer(Component component) {
	this.component = component;

	// Set up graphics configuration (ie. screen + visual):

	config = (XGraphicsConfiguration)
	    component.getGraphicsConfiguration();

	if (config == null) {
	    // This will usually only happen for toplevel windows
	    config = getXToolkit().getDefaultXGraphicsConfiguration();
	} 

	Rectangle bounds = component.getBounds();
	parent = locateParentWindow(bounds);

	
	// Windows in X must atleast be of size 1x1
	boolean boundsChanged = false;
	if (bounds.width < 1) {
	    boundsChanged = true;
	    bounds.width = 1;
	}
	if (bounds.height < 1) {
	    boundsChanged = true;
	    bounds.height = 1;
	}
	
	/* don't worry about this pinging back to us, since the real
	   component object has not yet received a reference to this
	   peer object. */
	component.setBounds(bounds);
	
	    
	no.rwr.x.WindowAttributes attributes = new WindowAttributes();

	/* Set background color */
	Color bg = component.getBackground();
	if (bg != null) {
	    int[] components = {
		bg.getRed(),
		bg.getGreen(),
		bg.getBlue(),
		0xff
	    };

	    ColorModel cm = config.getColorModel();
	    long pixel = cm.getDataElement(components, 0);
	    attributes.setBackground(pixel);
	};
	
	/* Set exposure mask so that we get exposure events
	   that can be translated into paint() calls. */
	long eventMask = WindowAttributes.MASK_EXPOSURE;

	/* It would be nice to set up all other required events here,
	   but it is not possible to do so before after all the
	   children of this component has been realized. The reason is
	   that it is not determined whether a componen is lightweight
	   before after the addNotify() method has been called. Thus,
	   it is not possible for parent component to determine what
	   events it needs to furnish for lightweight
	   children. Instead, we currently rely on the component
	   calling our setEventMask() method after the correct event
	   mask has been determined. */

	attributes.setEventMask(eventMask);
	
	    
	// set some attributes?

	/* don't allow event queue to process events from the newly
	   created window before this peer has been registered as
	   client data. */
	synchronized (getXToolkit().queue) {
	    window = new no.rwr.x.Window(parent, bounds, attributes);
	    window.setClientData(this); /* make it possible to find back
					   to this peer object. Used by
					   XEventQueue. */
	}
	
	initWindowProperties();
	
	if (component.isVisible()) {
	    window.map();
	    ensureFlush();
	} 
    }

    /** Override this in subclasses to implement other ways of
	obtaining parent windows. Toplevel windows will typically have
	a different implementation. */
    no.rwr.x.Window locateParentWindow(Rectangle bounds) {
	Container parent = component.getParent();
	while (parent.isLightweight()) {
	    bounds.x += parent.getX();
	    bounds.y += parent.getY();
	    parent = parent.getParent();
	    // a null pointer here is a genuine error
	}

	XCanvasPeer parentPeer = (XCanvasPeer) parent.getPeer();
	if (parentPeer == null)
	    throw new NullPointerException("Parent has no peer. This should " +
					   "not be possible, since the " +
					   "calls leading here should come " +
					   "from parent, after it has " +
					   "that the parent peer is set.");
	return parentPeer.window;
    }
    

    /** Template method to allow subclasses to apply properties
	to X11 window right after creation. */
    void initWindowProperties() {
    }
    

    XToolkit getXToolkit() {
	return XToolkit.INSTANCE;
    }

    protected void ensureFlush() {
	getXToolkit().flushIfIdle();
    }
    

    public GraphicsConfiguration getGraphicsConfiguration() {
	return config;
    }

    public Component getComponent() {
	return component;
    }


    
    // -------- java.awt.peer.ComponentPeer implementation

    public Toolkit getToolkit() {
	return getXToolkit();
    }

    public Graphics getGraphics() {
	return new XGraphics(window, config, component.getBackground());
    }

    public Dimension getMinimumSize () {
	return MIN_SIZE;
    }

    public Dimension getPreferredSize () {
	return component.getSize();
    }
    
    public void setBounds(int x, int y, int width, int height) {
	window.setBounds(x, y, width, height);
    }
    
    public void paint(Graphics gfx) {
	// do nothing by default
    }
    
    public void setVisible(boolean visible) {
	if (visible) {
	    window.map();
	    ensureFlush();	    
	} else {
	    throw new UnsupportedOperationException("unmap not implemented");
	}
    }
	
    public void setForeground(Color color) {
	/* default canvas peer does keep track of font, since it won't
	   paint anything. */
    }
    public void setBackground(Color color) {
	throw new UnsupportedOperationException("not implemented");
    }
	
    public void setFont(Font font) {
	/* default canvas peer does keep track of font, since it won't
	   write anything. */
    }

    public void setEventMask(long eventMask) {
	no.rwr.x.WindowAttributes attributes = new WindowAttributes();

	long xEventMask = WindowAttributes.MASK_EXPOSURE;
	
	if ((eventMask & EventConstants.MOUSE_EVENT_MASK) != 0) {
	    xEventMask |=
		WindowAttributes.MASK_BUTTON_PRESS |
		WindowAttributes.MASK_BUTTON_RELEASE;
	}
	    
	attributes.setEventMask(xEventMask);
	window.setAttributes(attributes);
	ensureFlush();
    }



    public void repaint(int x, int y, int w, int h) {
	/* TODO?

	   X allows intelligent X servers to do smart
	   refreshing. Perhaps involve X in repainting of components,
	   rather that keeping it all within the local event queue. */

	PaintEvent updateEvent = new PaintEvent(component,
						PaintEvent.UPDATE,
						new Rectangle(x, y, w, h));
	getXToolkit().queue.postEvent(updateEvent);
    }

    public void processEvent(AWTEvent event) {
	if (event instanceof PaintEvent) {
	    PaintEvent pEvt = (PaintEvent) event;
	    Graphics gfx = getGraphics();
	    Shape clip = pEvt.getUpdateRect();
	    gfx.setClip(clip);

	    switch (event.getID()) {
	    case PaintEvent.PAINT:
		component.paint(gfx);
		break;
	    case PaintEvent.UPDATE:
		

		component.update(gfx);
		break;
	    default:
		throw new RuntimeException("unknown paint event");
	    }
	    gfx.dispose();
	} 
    }
}
