package no.rwr.awt.x;

import java.awt.*;

import no.rwr.x.Display;
import no.rwr.x.XAnyEvent;
import no.rwr.x.XExposeEvent;
import no.rwr.x.XButtonEvent;
import java.awt.event.PaintEvent;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.util.Vector;

import no.rwr.awt.LightweightRedirector;

public class XEventLoop implements Runnable {
    
    

    boolean CAUTIOUS_CONCURRENT_XLIB = false;
    final static int DEBUG_LEVEL = 0;


    Display display;
    EventQueue queue;
    XAnyEvent anyEvent;
    Thread awtThread;

    LightweightRedirector lightweightRedirector = new LightweightRedirector();
    
    public XEventLoop(Display display, EventQueue queue) {
	this.display = display;
	this.queue = queue;

	anyEvent = new XAnyEvent(display);
	awtThread = new Thread(this, "AWT thread for XEventQueue");
	awtThread.start();
    }

    public void run() {
	while (true) {
	    postNextEvent();
	}
    }
    

    void postNextEvent() {
	try {
	    AWTEvent evt = getNextEvent();

	    queue.postEvent(evt);
	} catch (InterruptedException ie) {
	    // FIXME: what now?
	    System.err.println(ie);
	}
    }

    
    /** get next event. Will block until events become available. */
 
    public AWTEvent getNextEvent() {
	// ASSERT:
	if (isIdle()) throw new Error("should not be idle");

	loadNextEvent();
	AWTEvent event = createEvent();
	
	event = lightweightRedirector.redirect(event);

	return event;
    }

    void loadNextEvent() {
	try {
	    setIdle(true);
	    
	    /* The code below will result in an XFlush(). However,
	       while we are waiting for events after calling XFlush(),
	       new X requests issued on other threads will not
	       automatically be flushed. This can lead to a deadlock
	       since XFlush() will not be called before after the
	       processing of the next event, and new events arriving
	       might be dependent on the delivery of the X
	       requests. 
	       
	       Code that does issues X requests should therefore call
	       flushIfIdle() after they're done, to ensure that the
	       requests are delivered in a timely manner. XFlush is
	       not run if event processing is underway, since we are
	       assured that the event loop execution will return to
	       this point, where requests are flushed again, before
	       waiting for new events.  */
	    
	    display.flush();
	    if (CAUTIOUS_CONCURRENT_XLIB) display.waitForEvents();

	    anyEvent.loadNext();
	    display.flush();
	} catch (RuntimeException re) {
	    System.err.println("Exception thrown on event thread:" + re);
	} finally {
	    setIdle(false);
	}
    }
    

    
    
    AWTEvent createEvent() {
	
	/* avoid attempting to get client data before client data has
	   been set. */
	Object peer;
	synchronized (this) {
	    peer = anyEvent.getWindow().getClientData();
	}
	
	    
	Component source = null;

	// Try to identify source component
	
	if (peer instanceof XCanvasPeer) {
	    source = ((XCanvasPeer) peer).getComponent();
	}
	
	if (source == null) {
	    String msg = "unable to locate source for event (" +
		anyEvent + ")";
	    throw new RuntimeException(msg);
	}

	/* if a mapping from anyEvent to AWTEvent is possible, 
	   construct a new AWTEvent and return it. */
	
	int type = anyEvent.getType();
	switch (type) {
	    
	case XAnyEvent.TYPE_EXPOSE:
	    return createPaintEvent(source);
	case XAnyEvent.TYPE_BUTTON_PRESS:
	case XAnyEvent.TYPE_BUTTON_RELEASE:
	    return createMouseEvent(type, source);
	default:
	    String msg = "Do no know how to handle event (" + anyEvent + ")";
	    throw new RuntimeException(msg);
	}
	
    }
    
    AWTEvent createPaintEvent(Component src) {
	    XExposeEvent expose = new XExposeEvent(anyEvent);
	    PaintEvent pe = new PaintEvent(src, PaintEvent.PAINT,
					   expose.getBounds());
	    return pe;
    }
    


    /* Warning: this makes assumptions on the contents of X.h... 
       Button1 = 1, Button2 = 2, etc... */
    static final int[] buttonToModifierMap = {
	0,
	InputEvent.BUTTON1_MASK,
	InputEvent.BUTTON2_MASK,
	InputEvent.BUTTON3_MASK
    };


    AWTEvent createMouseEvent(int type, Component src) {    
	XButtonEvent buttonEvt = new XButtonEvent(anyEvent);

	int modifiers = buttonToModifierMap[buttonEvt.button];
	int state = buttonEvt.state;
	
	// remap bits from state to modifiers:
	
	if ((state & XButtonEvent.MASK_SHIFT) != 0)
	    modifiers |= InputEvent.SHIFT_MASK;
	
	
	if ((state & XButtonEvent.MASK_CONTROL) != 0)
	    modifiers |= InputEvent.CTRL_MASK;
	
	
	/* FIXME: we need additional X code to properly map
	   MODn states to input modifiers */
	
	int clickCount = 1; // FIXME... Can't get this from X.
	boolean popupTrigger = false; // FIXME: look up policy somewhere
	
	int x = buttonEvt.x;
	int y = buttonEvt.y;

	int id = (type == XAnyEvent.TYPE_BUTTON_PRESS) ?
	    MouseEvent.MOUSE_PRESSED :
	    MouseEvent.MOUSE_RELEASED;
	
	MouseEvent me = new MouseEvent(src,
				       id,
				       buttonEvt.time, modifiers,
				       buttonEvt.x, buttonEvt.y,
				       clickCount, popupTrigger);
	return me;
    }

    public void flushIfIdle() {
	if (isIdle()) {
	    if (DEBUG_LEVEL>0) System.out.println("flushing (idle)");
	    display.flush();
	} else {
	    if (DEBUG_LEVEL>0) System.out.println("not flushing");
	}
    }

    volatile boolean idle = false;

    final synchronized void setIdle(boolean idle) {
	if (DEBUG_LEVEL>0) System.out.println("idle=" + idle);
	this.idle = idle;
    }

    final synchronized boolean isIdle() {
	return idle;
    }
}
