package no.rwr.image;
import java.awt.image.*;
import no.rwr.util.*;

public class IntFieldPacker implements RasterOp {

    static final int SOURCE_BIT_WIDTH = 8;
    static final int LUT_SIZE = 1 << SOURCE_BIT_WIDTH;
    static final int LUT_MASK = LUT_SIZE - 1;
    
    static final int NUM_BANDS = 3;

    int[][] lut = new int[NUM_BANDS][LUT_SIZE];
    
    public IntFieldPacker(ComponentSampleModel srcCSM,
			  PackedColorModel destPCM) {

	// Precalculate LUTs:

	// FIXME: Cache precalculated LUTs?
	
	int[] bandOffsets = srcCSM.getBandOffsets();
	int[] masks = destPCM.getMasks();

	int[] maskOrder = ArrayUtil.invert(bandOffsets, null);
	
	if (!ArrayUtil.isCompleteInverse(bandOffsets, maskOrder)) { // ASSERT
	    throw new IllegalArgumentException("bands are not " +   // ASSERT
					       "tightly packed");   // ASSERT
	}                                                           // ASSERT
	
	BitMaskExtent extent = new BitMaskExtent();

	for (int b=0; b<NUM_BANDS; b++) {
	    int[] bLUT = lut[b];
	    int mask = masks[maskOrder[b]];
	    extent.setMask(mask);
	    int shift = extent.leastSignificantBit + extent.bitWidth 
		- SOURCE_BIT_WIDTH;

	    if (shift>0) {
		for (int i=0; i<LUT_SIZE; i++)
		    bLUT[i] = (int) ((i<<shift) & mask);
	    } else if (shift<0) {
		shift = -shift;
		for (int i=0; i<LUT_SIZE; i++) 
		    bLUT[i] = (int) ((i>>>shift) & mask);
	    } else {
		for (int i=0; i<LUT_SIZE; i++)
		    bLUT[i] = (int) (i & mask);
	    }
	}
    }

    public WritableRaster filter(Raster src, WritableRaster dest) {
    	if (dest == null) dest = createCompatibleDestRaster(src);

	DataBufferByte srcDB = (DataBufferByte) src.getDataBuffer();
	DataBufferInt  destDB = (DataBufferInt)  dest.getDataBuffer();

	ComponentSampleModel  srcSM =
	    (ComponentSampleModel) src.getSampleModel();

	SinglePixelPackedSampleModel destSM =
	    (SinglePixelPackedSampleModel) dest.getSampleModel();
	
	// Calculate offset to data in the underlying arrays

	int  srcScanlineStride =  srcSM.getScanlineStride();
	int destScanlineStride =  destSM.getScanlineStride();

	int  srcX =  src.getMinX() -  src.getSampleModelTranslateX();
	int  srcY =  src.getMinY() -  src.getSampleModelTranslateY();
	int destX = dest.getMinX() - dest.getSampleModelTranslateX();
	int destY = dest.getMinY() - dest.getSampleModelTranslateY();

	int numBands = srcSM.getNumBands();

	/* We can't use getOffset(x, y) from the sample model since we
	   don't want the band offset added in. */
	
	int srcOffset = 
	    numBands*srcX + srcScanlineStride*srcY +    // from sample model
	    srcDB.getOffset();                          // from data buffer

	int destOffset = 
	    srcX + srcScanlineStride*srcY +             // from sample model
	    srcDB.getOffset();                          // from data buffer


	// Determine how much, and how many segments to extract.

	int rowSize = src.getWidth();
	int h = src.getHeight();

	if (((rowSize*numBands) == srcScanlineStride) &&
	    (rowSize == destScanlineStride)) {
	    // collapse scan line blits to one large blit.
	    //System.out.println("scan lines sweeps collapsed"); // DEBUG
	    
	    rowSize *= h;
	    h = 1;
	}


	
	// Do blitting

	byte[]   srcArray =  srcDB.getData();
	int[] destArray = destDB.getData();

	int[] lut0 = lut[0];
	int[] lut1 = lut[1];
	int[] lut2 = lut[2];

	if ((destArray == null) || (srcArray == null) ||
	    (lut0 == null) || (lut1 == null) || (lut2 == null)) {
	    throw new NullPointerException("unexpected null-pointers");
	}
	

	for (int y = 0; y<h; y++) {
	    int sOffset = srcOffset;
	    int dOffset = destOffset;

	    for (int x = 0; x<rowSize; x++) {		
		destArray[dOffset++] = (int) (
		    lut0[srcArray[sOffset++] & LUT_MASK] |
		    lut1[srcArray[sOffset++] & LUT_MASK] |
		    lut2[srcArray[sOffset++] & LUT_MASK]);
	    }
	    srcOffset  +=  srcScanlineStride;
	    destOffset += destScanlineStride;
	}

	return dest;
    }

    public WritableRaster createCompatibleDestRaster(Raster src) {

	/* FIXME: Maybe we should explicitly create a raster with a
	   tightly pixel packed sample model, rather than assuming
	   that the createCompatibleWritableRaster() method in Raster
	   will create one. */

	return src.createCompatibleWritableRaster();
    }
}
